#!/usr/bin/python
"""
Chronicle - Scheduled event manager.
    Deals with appointments, reminders, anniversaries, etc.

Copyright (C) 2001 James ''Wez''Weatherall

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

ToDo:
1)  Rewrite database to use shelve?
    Currently, we rewrite the whole thing whenever stuff changes.
    With shelve/dbm, we would only rewrite changed stuff, I think.
    This would be better for the FLASH RAM.

"""

try: import pylau
except ImportError: pass

import gtk
try: import cPickle
except:
    import pickle
    cPickle = pickle
import sys
import os
import time
import socket
import string
import threading

if os.name == "posix":

    # - Support for the iPAQ AT command under Familiar
    if ("armv4l" in os.uname()) and os.popen("ipkg status atd").read():
        # *** Hack - how do we really detect Familiar?
        _has_at = 1
    else:
        _has_at = 0

    if _has_at:
        _atdir = "/var/spool/at"
        
        def _set_at(when):
            # - Create the AT entry
            fd = os.open("/var/spool/at/%s.%s" %
                         (int(time.mktime(when)), os.getpid()),
                         os.O_WRONLY | os.O_CREAT,
                         0755)
            try:
                os.write(fd,"#!/bin/sh\necho 1 >/tmp/.schedule-trigger\nrm -f $0\n")
            finally:
                os.close(fd)
            open("/var/spool/at/trigger", "w").write("\n")

        def _del_at():
            atcmds = os.listdir(_atdir)
            for at in atcmds:
                if os.path.splitext(at)[1] == str("."+str(os.getpid())):
                    print "Unlinking AT command"
                    os.unlink(os.path.join(_atdir, at))

else:
    _has_at = 0

if not _has_at: print "schedule: operating without at"

def time_now():
    return tuple(time.localtime(time.time()))

def time_correct(tval):
    return apply(time_mk, tval[0:6])

def time_mk(year, month, day, hour, minute, sec):
    return tuple(time.localtime(time.mktime((year, month, day, hour, minute, sec, 0, 0, -1))))

def time_add(tval, secs):
    return tuple(time.localtime(time.mktime(tval)+secs))

def date_str(tval):
    return time.strftime("%a, %d %b %Y", tval)

def date_same(d1, d2):
    return d1[0:3] == d2[0:3]

def date_add(d1, d2):
    return time_mk(d1[0]+d2[0], d1[1]+d2[1], d1[2]+d2[2], 0, 0, 0)

def date_now():
    return apply(time_mk, time_now()[:3]+(0,0,0))

def time_str(tval):
    return time.strftime("%H:%M", tval)

class EventMgr:
    def __init__(self, filename):
        self.__setstate__({"fields":Event.Fields, "events":[]})
        self.filename = filename
    def new_event(self):
        return Event(self)
    def add_event(self, event):
        event.check_repeat()
        if event not in self.list:
            self.modified = 1
            self.list.append(event)
    def del_event(self, event):
        if event in self.list:
            self.modified = 1
            self.list.remove(event)
    def sort(self):
        for e in self.list: e.check_repeat()
        self.list.sort(self.cmp_events)
    def cmp_events(self, a, b):
        if a.start < b.start:
            return -1
        else:
            return 0
    def dump(self):
        if not self.timer:
            # - Start the commit timer
            self.timer = gtk.timeout_add(30000, self.on_dump_timeout)
    def dump_now(self):
        # - Do we need to dump the database?
        do_dump = self.modified
        self.modified = 0
        for event in self.list:
            if event.modified:
                do_dump = 1
                event.modified = 0
        if do_dump:
            print "schedule:committed database"
            cPickle.dump(self, open(self.filename, "wb"))
        return do_dump
    def on_dump_timeout(self):
        done = self.dump_now()
        if not done: self.timer = None
        return done
    def set_context(self, context):
        self.context = context
    def __setstate__(self, state):
        self.timer = None
        self.modified = 0
        self.list=[]
        fields = state["fields"]
        for item in state["events"]:
            event = Event(self)
            for i in range(len(fields)):
                setattr(event, fields[i], item[i])
            self.list.append(event)
    def __getstate__(self):
        def f(t):
            l = []
            for name in Event.Fields:
                l.append(getattr(t, name))
            return l
        result = {"fields":Event.Fields,
                  "events":map(f, self.list)}
        return result

class Event:
    Fields = ["text", "start", "timed", "duration", "audible", "repeat_type"]
    def __init__(self, mgr):
        self.mgr = mgr
        self.modified = 0
        
        self.text = ""
        self.start = time_now()
        self.timed = 0
        self.duration = 0
        self.audible = 0

        self.repeat_type = ""
    def dump(self):
        self.mgr.dump()
    def dump_now(self):
        self.mgr.dump_now()
    def check_repeat(self):
        while time.localtime(time.time())[0:3] > self.start[0:3]:
            if self.repeat_type == "D":
                self.start=time_correct(self.start[:2]+(self.start[2]+1,)+self.start[3:])
            elif self.repeat_type == "M":
                self.start=time_correct(self.start[:1]+(self.start[1]+1,)+self.start[2:])
            elif self.repeat_type == "Y":
                self.start=time_correct((self.start[0]+1,)+self.start[1:])
            elif self.repeat_type == "W":
                weekday = self.start[6]
                while 1:
                    self.start=time_correct(self.start[:2]+(self.start[2]+1,)+self.start[3:])
                    if self.start[6] == weekday:
                        break
            else:
                break
    def __setstate__(self, state):
        self.__dict__ = state
        self.modified = 0
    def __getstate__(self):
        state = {}
        for field in Event.Fields:
            state[field] = getattr(self, field)
        return state

def HButtonBox(*buttons):
    box=gtk.GtkHButtonBox()
    box.set_spacing(0)
    box.set_child_ipadding_default(0, 0)
    box.set_child_size_default(0,0)
    for i in buttons:
        box.add(i)
    return box

def Button(title, func, name=None):
    b=gtk.GtkButton(title)
    if name: b.set_name(name)
    b.connect("clicked", func)
    b.show()
    return b

def Frame(title, content):
    f=gtk.GtkFrame(title)
    f.add(content)
    return f

def MenuItem(title, func, data=None, name=None):
    i=gtk.GtkMenuItem(title)
    if func: i.connect("activate", func)
    if name: i.set_name(name)
    if data: i.set_data(data[0], data[1])
    i.show()
    return i

def Menu(items):
    m=gtk.GtkMenu()
    for i in items:
        m.add(i)
    return m

# *** Supposed to produce better behaviour with xstroke - doesn't seem to work
def gdk_intern_atom(name, only_if_exists=1):
    import _gtk
    atom = _gtk.gdk_atom_intern(name, only_if_exists)
    if not int(atom): print "warning: unable to intern", name
    return atom

def gdk_set_win_prop(window, property, type,
                     format, mode, data, only_if_exists=1):
    import _gtk, GDK
    atom = gdk_intern_atom(property, only_if_exists)
    if not int(atom): return 0
    int_atom = gdk_intern_atom(type, only_if_exists)
    if not int(int_atom): return 0
    if mode=="REPLACE":gdk_mode = GDK.PROP_MODE_REPLACE
    elif mode=="APPEND":gdk_mode = GDK.PROP_MODE_APPEND
    elif mode=="PREPEND":gdk_mode = GDK.PROP_MODE_PREPEND
    try:
        window.property_change(atom, int_atom, format, gdk_mode, data)
        return 1
    except:
        return 0

def gdk_set_recursive_win_prop(window, property, type,
                               format, mode, data, only_if_exists=1):
    children = window.children
    if children:
        for child in children:
            gdk_set_recursive_win_prop(child, property, type, format, mode, data, only_if_exists)
    else:
        gdk_set_win_prop(window, property, type, format, mode, data, only_if_exists)
# ***

class GtkTimeEntry(gtk.GtkFrame):
    def __init__(self, title, hour, minute, second):
        gtk.GtkFrame.__init__(self, title)

        # - Time buttons
        adj = gtk.GtkAdjustment(hour, 0, 23, 1,1,1)
        self.hour = gtk.GtkSpinButton(adj,1,0)
        adj = gtk.GtkAdjustment(minute, 0, 59, 15,15,1)
        self.minute = gtk.GtkSpinButton(adj,15,0)
        second = 0
        adj = gtk.GtkAdjustment(second, 0, 59, 15,15,1)
        self.second = gtk.GtkSpinButton(adj,15,0)

        hbox = gtk.GtkHBox()
        hbox.pack_start(self.hour)
        hbox.pack_start(self.minute)
        #hbox.pack_start(self.second)
        self.add(hbox)
        self.show_all()

class Reminder(gtk.GtkWindow):
    def __init__(self, event, mgrview=None):
        gtk.GtkWindow.__init__(self)
        self.set_title("Reminder")
        self.connect("destroy", self.OnDestroy)

        box = gtk.GtkVBox()

        title = gtk.GtkLabel("Reminder %s %s" % (date_str(event.start), time_str(event.start)))
        box.pack_start(title, gtk.FALSE)

        text_hbox = gtk.GtkHBox()
        box.pack_start(text_hbox)
        vadj = gtk.GtkAdjustment()
        text = gtk.GtkText(None, vadj)
        text_hbox.pack_start(text)
        text.set_editable(0)
        vscrollbar = gtk.GtkVScrollbar(vadj)
        text_hbox.pack_start(vscrollbar, gtk.FALSE)

        button_box = gtk.GtkHBox()
        close = gtk.GtkButton("Close")
        close.connect("clicked", self.destroy)
        if not event.repeat_type:
            remove = gtk.GtkButton("Close & Destroy")
            remove.connect("clicked", self.OnRemoveClicked)
            button_box.pack_start(remove, gtk.TRUE, gtk.FALSE)
        button_box.pack_start(close, gtk.TRUE, gtk.FALSE)
        box.pack_start(button_box, gtk.FALSE)
        self.add(box)
        text.insert(None, None, None, event.text)
        self.show_all()

        self.__event = event
        self.__mgrview = mgrview

        if event.audible: self.timer = gtk.timeout_add(3000, self.OnTimeout)
        else: self.timer = None
    def OnRemoveClicked(self, widget):
        self.__event.mgr.del_event(self.__event)
        if self.__mgrview: self.__mgrview.Refresh()
        self.destroy()
    def OnDestroy(self, widget):
        if self.timer: gtk.timeout_remove(self.timer)
    def OnTimeout(self):
        cmd = "madplay /usr/lib/media/tada.mp3 >/dev/null 2>/dev/null &"
        os.system(cmd)
        print "reminder:Played sample"
        return 1

class EventView(gtk.GtkWindow):
    def __init__(self, event):
        gtk.GtkWindow.__init__(self)
        self.event = event
        
        self.set_title("Event")
        self.set_default_size(240,320)

        # - Button bar
        buttons=HButtonBox(Button("Save", self.OnSaveClicked, "Commit"),
                           Button("Cancel", self.OnCancelClicked, "Discard"))

        # - Event text
        self.text = gtk.GtkText()
        self.text.set_editable(1)
        self.text.set_usize(32,32)
        self.text.insert_defaults(self.event.text)

        # - Event (start) date
        self.cal = gtk.GtkCalendar()
        tv = self.event.start
        self.cal.select_month(tv[1]-1,tv[0])
        self.cal.select_day(tv[2])
        self.cal.show()

        # - Start time and duration
        tv = self.event.start
        self.timed = gtk.GtkCheckButton()
        self.timed.connect("clicked", self.OnStartClicked)
        self.start = GtkTimeEntry("Start Time", tv[3], tv[4], tv[5])
        startbox = gtk.GtkHBox()
        startbox.pack_start(self.timed, 0, 0)
        startbox.pack_end(self.start)

        # - Duration
        diff=self.event.duration
        self.duration_set= gtk.GtkCheckButton()
        self.duration_set.connect("clicked", self.OnDiffClicked)
        self.duration = GtkTimeEntry("Duration",
                                diff / 3600,
                                (diff/60) % 60,
                                diff % 60)
        diffbox = gtk.GtkHBox()
        diffbox.pack_start(self.duration_set, 0, 0)
        diffbox.pack_end(self.duration)
        
        timebox=gtk.GtkHBox()
        timebox.pack_start(startbox)
        timebox.pack_end(diffbox)

        # - Make start time & duration optional
        self.duration_set.set_active(self.event.duration > 0)
        self.OnStartClicked(self.timed)
        self.timed.set_active(self.event.timed)
        self.OnDiffClicked(self.duration_set)

        # - Audible alarm
        ar_box = gtk.GtkHBox()
        self.audible = gtk.GtkCheckButton("Audible alarm")
        self.audible.set_active(self.event.audible)
        ar_box.pack_start(self.audible)

        # - Repeated appointments
        repeat = gtk.GtkOptionMenu()
        repeats=[MenuItem("No Repeat", self.OnRepeatSet,
                          ("repeat", None)),
                 MenuItem("Daily", self.OnRepeatSet,
                          ("repeat", "D")),
                 MenuItem("Weekly", self.OnRepeatSet,
                          ("repeat", "W")),
                 MenuItem("Monthly", self.OnRepeatSet,
                          ("repeat", "M")),
                 MenuItem("Yearly", self.OnRepeatSet,
                          ("repeat", "Y"))]
        for r in repeats:
            if r.get_data("repeat") == self.event.repeat_type:
                repeats.remove(r)
                repeats.insert(0, r)
                break
        repeat.set_menu(Menu(repeats))
        self.repeat_type = self.event.repeat_type
        ar_box.pack_end(repeat)

        # - Pack it all up
        vbox = gtk.GtkVBox()
        vbox.pack_start(Frame("Text", self.text))
        vbox.pack_start(self.cal)
        vbox.pack_start(timebox)
        vbox.pack_start(ar_box)
        vbox.pack_end(buttons, 0, 0)
        self.add(vbox)

        self.show_all()
        gdk_set_recursive_win_prop(self.cal.get_window(),
                                   "RECOGNIZE_PROTOCOLS",
                                   "ATOM", 32, "REPLACE",
                                   [gdk_intern_atom("RECOGNIZE_KEY",0)], 0)
        self.text.grab_focus()

        # - Handling repeat appointment settings
    def OnRepeatSet(self, widget):
        print "Set data"
        self.repeat_type = widget.get_data("repeat")

        # - Handle disabling of start time & duration
    def OnStartClicked(self, widget):
        active = self.timed.get_active()
        self.start.set_sensitive(active)
        self.duration_set.set_sensitive(active)
        self.duration.set_sensitive(active and self.duration_set.get_active())
    def OnDiffClicked(self, widget):
        active = self.duration_set.get_active()
        self.duration.set_sensitive(active)

        # - Save the event
    def OnSaveClicked(self, widget):
        if self.event.timed != self.timed.get_active():
            self.event.modified = 1
            self.event.timed = self.timed.get_active()
        if self.event.text != self.text.get_chars(0,-1):
            self.event.modified = 1
            self.event.text = self.text.get_chars(0,-1)
        date = self.cal.get_date()
        if self.event.timed:
            start = time_mk(date[0], date[1]+1, date[2],
                            self.start.hour.get_value_as_int(),
                            self.start.minute.get_value_as_int(),
                            self.start.second.get_value_as_int())
        else:
            start = time_mk(date[0], date[1]+1, date[2],0,0,0)
        if self.event.start != start:
            self.event.modified = 1
            self.event.start = start
        if self.duration_set.get_active():
            duration = (self.duration.hour.get_value_as_int() * 3600 +
                        self.duration.minute.get_value_as_int() * 60 +
                        self.duration.second.get_value_as_int())
        else:
            duration = 0
        if self.event.duration != duration:
            self.event.modified = 1
            self.event.duration = duration
        if self.event.audible != self.audible:
            self.event.audible = self.audible.get_active()
            self.event.modified = 1
        if self.event.repeat_type != self.repeat_type:
            self.event.repeat_type = self.repeat_type
            self.event.modified = 1
        self.event.mgr.add_event(self.event)
        self.event.dump()
        self.destroy()
    def OnCancelClicked(self, widget):
        self.destroy()

_SigFile = "/tmp/.schedule-trigger"
class _SigThread(threading.Thread):
    def __init__(self, pipe):
        threading.Thread.__init__(self)
        self.__out = pipe
        self.__active = 1
    def run(self):
        print "sched:thread start"
        try: os.unlink(_SigFile)
        except: pass
        os.mkfifo(_SigFile)
        while self.__active:
            sigport = os.open(_SigFile, os.O_RDONLY)
            try:
                data = repr(os.read(sigport, 1024))
                if self.__out: os.write(self.__out, data)
            finally:
                os.close(sigport)
        try: os.unlink(_SigFile)
        except: pass
        print "sched:thread end"
    def close(self):
        try: os.close(self.__out)
        except: pass
        self.__out = None
        self.__active = 0
        open(_SigFile, "w").write("quit\n")
        self.join()

def _EventMgrFilter_weekplus(x):
    now = date_now()[:3]
    then = date_add(now, (0,0,7))[:3]
    start = x.start[:3]
    return (now <= start) and (start <= then)
def _EventMgrFilter_week(x):
    now = time_now()
    start = date_add(now, (0,0,-now[6]))[:3]
    end = date_add(now, (0,0,7-now[6]))[:3]
    return (start <= x.start[:3]) and (x.start[:3] <= end)

_EventMgrFilter = {
    "Future": lambda x: (time_now() <= x.start[0:6]),
    "All": lambda x:1,
    "Week": _EventMgrFilter_week,
    "Week+": _EventMgrFilter_weekplus,
    "Month": lambda x: (time_now()[1] == x.start[1]),
    "Past": lambda x: (time_now() > x.start[0:6]),
    "Today": lambda x: date_same(date_now(), x.start)}

class _EventMgrView_config:
    def __init__(self):
        self.view = "Future"
        self.alarm_time = time.time()

class EventMgrView(gtk.GtkWindow):

    def __init__(self, event_mgr):
        gtk.GtkWindow.__init__(self)
        self.mgr = event_mgr
        try: self.pconfig = cPickle.load(open("/tmp/.chronicle-config", "rb"))
        except: self.pconfig = _EventMgrView_config()
        self.alarm_timer = None
        self.alarm_pipe = None
        self.alarm_input_handler = None
        self.alarm_thread = None
        
        self.set_title("Chronicle")
        self.set_default_size(240,320)
        self.connect("destroy", self.OnDestroy)

        # - Menu bar
        menu_hbox = gtk.GtkHBox()
        menu = gtk.GtkMenuBar()
        file_item = gtk.GtkMenuItem("Chronicle")
        file_menu = gtk.GtkMenu()
        quit_item = gtk.GtkMenuItem("Quit")
        quit_item.connect("activate", self.destroy)
        file_menu.add(quit_item)
        file_item.set_submenu(file_menu)
        menu.add(file_item)

        # - View options
        view = gtk.GtkOptionMenu()
        items=[]
        for viewname in _EventMgrFilter.keys():
            item = MenuItem(viewname, self.OnViewSet, ("view", viewname))
            if viewname == self.pconfig.view: items.insert(0, item)
            else: items.append(item)
        view.set_menu(Menu(items))
    
        menu_hbox.pack_start(menu)
        menu_hbox.pack_end(view, 0, 0)

        # - Button bar
        buttons = HButtonBox(Button("New", self.OnNewClicked, "Create"),
                             Button("Edit", self.OnEditClicked),
                             Button("Destroy", self.OnDestroyClicked, "Destroy"),
                             Button("Quit", self.destroy))

        # - Create the list
        self.list = gtk.GtkCList(3,["Time", "F", "Event"])
        self.list.set_column_auto_resize(0, 1)
        self.list.set_column_auto_resize(1, 1)
        self.list.column_titles_show()
        self.list.connect("select-row", self.OnEventSelected)
        self.selected = None
        self.OnAlarmTimeout()

        # - Build the window layout
        vbox = gtk.GtkVBox()
        vbox.pack_start(menu_hbox, 0, 0)
        list_wnd = gtk.GtkScrolledWindow()
        list_wnd.add(self.list)
        list_wnd.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        vbox.add(list_wnd)
        vbox.pack_end(buttons, 0, 0)
        self.add(vbox)

        self.show_all()

    def OnDestroy(self, widget):
        self.mgr.dump_now()
        self.ClearAlarmTimeout()
        try: cPickle.dump(self.pconfig, open("/tmp/.chronicle-config", "wb"))
        except: print "warning: persistent view config not saved"
        gtk.mainquit()

    # - The bottom Button Bar
    def OnNewClicked(self, widget):
        event = self.mgr.new_event()
        event_view = EventView(event)
        event_view.connect("destroy", self.OnEventEditClosed, event_view)
    def OnEditClicked(self, widget):
        if self.selected:
            event_view = EventView(self.selected)
            event_view.connect("destroy", self.OnEventEditClosed, event_view)
    def OnDestroyClicked(self, widget):
        if self.selected:
            self.mgr.del_event(self.selected)
            self.selected = None
            self.Refresh()

    # - View modes
    def OnViewSet(self, widget):
        self.pconfig.view = widget.get_data("view")
        self.Refresh()

    # - ToDo item actions
    def OnEventEditClosed(self, widget, view):
        event = view.event
        self.selected = event
        self.Refresh(event)
    def OnEventSelected(self, widget, row, column, event):
        self.selected = self.list.get_row_data(row)
        if not self.selected: self.selected=None

    # - Redrawing the list
    def Refresh(self, event_refresh=None):
        # - Reset the alarm timeout
        self.SetAlarmTimeout()
        
        # - Sort and filter events
        self.mgr.sort()
        events = filter(_EventMgrFilter[self.pconfig.view], self.mgr.list)

        # - Strip out deleted or mis-ordered events and labels
        pos=0
        shown=[]
        while self.list.get_row_data(pos) != None:
            if self.list.get_row_data(pos) not in events:
                self.list.remove(pos)
            else:
                shown.append(self.list.get_row_data(pos))
                pos=pos+1

        # - Add in events in appropriate positions
        pos = 0
        now = time.localtime(time.time())
        for event in events:
            if event not in shown:
                while self.list.get_row_data(pos) != None:
                    # - Are we in the correct location?
                    current = self.list.get_row_data(pos)
                    if current and (current.start > event.start): break
                    pos = pos + 1

                # - Add a new entry
                if event.timed:
                    str_time = time_str(event.start)
                    if event.duration:
                        str_time = str_time + "-" + time_str(time_add(event.start, event.duration))
                else: str_time=""
                str_flags=""
                if event.audible:
                    str_flags = str_flags + "A"
                if event.repeat_type:
                    str_flags = str_flags + event.repeat_type
                self.list.insert(pos, [str_time, str_flags, event.text])
                if pos % 2: self.list.set_background(pos, gtk.GdkColor(-16000, -8000, -16000))
                else: self.list.set_background(pos, gtk.GdkColor(-8000, -8000, -16000))
                self.list.set_row_data(pos, event)
                shown.append(event)

        if self.selected:
            for child in self.list.children():
                if child.get_data("event") == self.selected:
                    child.select()

        # - Re-insert missing data labels for days with appointments
        date = None
        pos = 0
        while self.list.get_row_data(pos) != None:
            event = self.list.get_row_data(pos)
            if event:
                if (not date) or (not date_same(date, event.start)):
                    date = event.start[0:3]
                    day_lbl, date_lbl=string.split(date_str(time_correct(event.start[0:3]+(0, 0, 0, 0, 0, -1))), ",")
                    if date_same(now, date): day_lbl = "TODAY"
                    self.list.insert(pos, [day_lbl, "", date_lbl])
                    self.list.set_row_data(pos, 0)
                    self.list.set_selectable(pos, 0)
                    #self.list.set_background(pos, gtk.GdkColor(-12000, -8000, -12000))
                    self.list.set_background(pos, gtk.GdkColor(0, 0, 0))
                    self.list.set_foreground(pos, gtk.GdkColor(-1, -1,-1))
            pos=pos+1

        self.mgr.dump()

    def SetAlarmTimeout(self):
        # - Set the alarm timeout timer correctly
        self.ClearAlarmTimeout()
        next = None
        for event in self.mgr.list:
            when = time.mktime(event.start)
            if (self.pconfig.alarm_time < when):
                if not next or (event.start < next.start):
                    next = event

        if next:
            # - Set the timer
            diff = long(max(0.0, time.mktime(next.start) - time.time()))
            diff = min(3600L, diff)
            diff = diff*1000L
            self.alarm_timer = gtk.timeout_add(diff, self.OnAlarmTimeout)
            if next.audible:
                if _has_at:
                    # - iPAQ has an AT command we use to power it up
                    #   at a specific time.
                    # - We implement the audible alarm this way.
                    #   If an alarm is inaudible, we don't wake to show it!

                    # - Start the AT handling thread
                    self.alarm_pipe, outpipe = os.pipe()
                    self.alarm_input_handler = gtk.input_add(self.alarm_pipe,
                                                             gtk.GDK.INPUT_READ,
                                                             self.OnAlarmTimeout)
                    self.alarm_thread = _SigThread(outpipe)
                    self.alarm_thread.start()

                    _set_at(next.start)

    def ClearAlarmTimeout(self):
        if self.alarm_timer:
            gtk.timeout_remove(self.alarm_timer)
            self.alarm_timer = None
        if _has_at: _del_at()
        if self.alarm_thread:
            self.alarm_thread.close()
            self.alarm_thread = None
        if self.alarm_input_handler:
            gtk.input_remove(self.alarm_input_handler)
            self.alarm_input_handler = None
        if self.alarm_pipe:
            os.close(self.alarm_pipe)
            self.alarm_pipe = None
            
    def OnAlarmTimeout(self, *args):
        # - Show any pending reminders
        now = time.time()
        done = 0
        for event in self.mgr.list:
            when = time.mktime(event.start)
            if ((self.pconfig.alarm_time < when) and (when < now)):
                Reminder(event, self)
        self.pconfig.alarm_time = now
        self.Refresh()
        return 0

def Main(argv):
    if len(argv) >= 2:
        filename = argv[1]
    else:
        filename = "/tmp/schedule.pickle"
    try:
        # - Hack our way around the rename to Chronicle...
        sys.modules["schedule"] = sys.modules[__name__]
        event_mgr = cPickle.load(open(filename, "rb"))
        event_mgr.filename = filename
    except:
        print "warning: unable to load %s" % filename, sys.exc_info()[0]
        event_mgr = EventMgr(filename)
    event = EventMgrView(event_mgr)
    gtk.mainloop()

if __name__ == "__main__": Main(sys.argv)




